@@ -28,6 +28,8 @@ using ICSharpCode.SharpZipLib.Zip;
 using Monodoc.Lucene.Net.Index;
 using Monodoc.Lucene.Net.Documents;
 
+using Mono.Documentation;
+
 using BF = System.Reflection.BindingFlags;
 
 //
@@ -146,23 +148,33 @@ public static class EcmaDoc {
 // It populates a tree with contents during the help assembly
 //
 public class EcmaProvider : Provider {
-	string basedir;
+	ArrayList/*<string>*/ asm_dirs = new ArrayList ();
+
+	public EcmaProvider ()
+	{
+	}
 
 	public EcmaProvider (string base_directory)
 	{
-		basedir = base_directory;
+		AddDirectory (base_directory);
+	}
 
-		if (!Directory.Exists (basedir))
-			throw new FileNotFoundException (String.Format ("The directory `{0}' does not exist", basedir));
+	public void AddDirectory (string directory)
+	{
+		if (!Directory.Exists (directory))
+			throw new FileNotFoundException (String.Format ("The directory `{0}' does not exist", directory));
+		asm_dirs.Add (directory);
 	}
 	
 	public override void PopulateTree (Tree tree)
 	{
-		string [] namespaces = Directory.GetDirectories (basedir);
-		
+		ArrayList ns_dirs = new ArrayList ();
+		foreach (string asm in asm_dirs) {
+			ns_dirs.AddRange (Directory.GetDirectories (asm));
+		}
 
-		
-		foreach (string ns in namespaces){
+		foreach (string ns in ns_dirs){
+			string basedir = Directory.GetParent (ns).FullName;
 			string [] files = Directory.GetFiles (ns);
 			Node ns_node = null;
 			string tn = null;
@@ -184,10 +196,11 @@ public class EcmaProvider : Provider {
 					if (File.Exists (ns_summary_file)) {
 						XmlDocument nsSummaryFile = new XmlDocument ();
 						nsSummaryFile.Load (ns_summary_file);
+						namespace_realpath [tn] = ns_summary_file;
 						
 						XmlNode ns_summary = nsSummaryFile.SelectSingleNode ("Namespace/Docs/summary");
-						if (ns_summary != null && ns_summary.InnerText != "To be added." && ns_summary.InnerText != "") {
-							namespace_summaries [tn] = nsSummaryFile.SelectSingleNode ("Namespace/Docs/summary");
+						if (ns_summary != null && ns_summary.InnerText.Trim () != "To be added." && ns_summary.InnerText != "") {
+							namespace_summaries [tn] = ns_summary;
 							namespace_remarks [tn] = nsSummaryFile.SelectSingleNode ("Namespace/Docs/remarks");
 						}
 						
@@ -291,7 +304,7 @@ public class EcmaProvider : Provider {
 				e.AppendChild (copy);
 				elements.AppendChild (e);
 			}
-			hs.PackXml ("xml.summary." + ns, doc);
+			hs.PackXml ("xml.summary." + ns, doc,(string) namespace_realpath[ns]);
 		}
 		
 		
@@ -308,14 +321,45 @@ public class EcmaProvider : Provider {
 				summary.AppendChild (nsSummary.ImportNode (n,true));
 			else
 				summary.AppendChild (nsSummary.CreateElement("summary"));
-			
 		}
-		tree.HelpSource.PackXml ("mastersummary.xml", nsSummary);
+		tree.HelpSource.PackXml ("mastersummary.xml", nsSummary, null);
+		AddExtensionMethods (tree);
+	}
+
+	void AddExtensionMethods (Tree tree)
+	{
+		XmlDocument extensions = null;
+		XmlNode root = null;
+		int numMethods = 0;
+		foreach (string asm in asm_dirs) {
+			string overview_file = Path.Combine (asm, "index.xml");
+			if (File.Exists (overview_file)) {
+				XmlDocument overview = new XmlDocument ();
+				overview.Load (overview_file);
+				XmlNodeList e = overview.SelectNodes ("/Overview/ExtensionMethods/*");
+				if (e.Count > 0) {
+					if (extensions == null) {
+						extensions = new XmlDocument ();
+						root = extensions.CreateElement ("ExtensionMethods");
+						extensions.AppendChild (root);
+					}
+					foreach (XmlNode n in e) {
+						++numMethods;
+						root.AppendChild (extensions.ImportNode (n, true));
+					}
+				}
+			}
+		}
+		if (extensions != null) {
+			Console.Error.WriteLine ("Have {0} extension methods", numMethods);
+			tree.HelpSource.PackXml ("ExtensionMethods.xml", extensions, "ExtensionMethods.xml");
+		}
 	}
 	       
-	static Hashtable class_summaries = new Hashtable ();
-	static Hashtable namespace_summaries = new Hashtable ();
-	static Hashtable namespace_remarks = new Hashtable ();
+	Hashtable class_summaries = new Hashtable ();
+	Hashtable namespace_summaries = new Hashtable ();
+	Hashtable namespace_remarks = new Hashtable ();
+	Hashtable namespace_realpath = new Hashtable ();
 	XmlDocument doc;
 	
 	void PopulateClass (string ns, Node ns_node, string file)
@@ -499,7 +543,7 @@ public class EcmaHelpSource : HelpSource {
 			if (css_ecma != null)
 				return css_ecma;
 			if (use_css) {
-				System.Reflection.Assembly assembly = System.Reflection.Assembly.GetCallingAssembly ();
+				System.Reflection.Assembly assembly = typeof(EcmaHelpSource).Assembly;
 				Stream str_css = assembly.GetManifestResourceStream ("mono-ecma.css");
 				css_ecma = (new StreamReader (str_css)).ReadToEnd();
 			} else {
@@ -508,6 +552,22 @@ public class EcmaHelpSource : HelpSource {
 			return css_ecma;
 		}
 	}
+	static string js;
+	public static string js_code {
+		get {
+			if (js != null)
+				return js;
+			if (use_css) {
+				System.Reflection.Assembly assembly = typeof(EcmaHelpSource).Assembly;
+				Stream str_js = assembly.GetManifestResourceStream ("mono-ecma-css.js");
+				js = (new StreamReader (str_js)).ReadToEnd();
+			} else {
+				js = String.Empty;
+			}
+			return js;
+		}
+	}
+
 	public override string GetText (string url, out Node match_node)
 	{
 		match_node = null;
@@ -522,12 +582,12 @@ public class EcmaHelpSource : HelpSource {
 			args.AddExtensionObject("monodoc:///extensions", ExtObject);
 			args.AddParam("show", "", "masteroverview");
 			string s = Htmlize(new XPathDocument (summary), args);
-			return BuildHtml (css_ecma_code, s); 
+			return BuildHtml (css_ecma_code, js_code, s); 
 		}
 		
 		if (url.StartsWith ("ecma:")) {
 			string s = GetTextFromUrl (url);
-			return BuildHtml (css_ecma_code, s); 
+			return BuildHtml (css_ecma_code, js_code, s); 
 		}
 
 		return null;
@@ -583,20 +643,23 @@ public class EcmaHelpSource : HelpSource {
 			
 			foreach (Node m in x.Nodes) {
 				string caption = m.Caption;
+				string ecaption = ToEscapedMemberName (caption);
 				if (m.IsLeaf) {
 					// No overloading (usually), is just the member name.  The whole thing for constructors.
-					if (caption == membername || caption == member) {
+					if (caption == membername || caption == member ||
+							ecaption == membername || ecaption == member) {
 						type_node = m;
 						return GetTextFromUrl (m.URL);
 					}
-				} else if (caption == member) {
+				} else if (caption == member || ecaption == member) {
 					// Though there are overloads, no arguments are in the url, so use this base node
 					type_node = m;
 					return GetTextFromUrl (m.URL);
 				} else {
 					// Check subnodes which are the overloads -- must match signature
 					foreach (Node mm in m.Nodes) {
-						if (mm.Caption == member) {
+						ecaption = ToEscapedTypeName (mm.Caption);
+						if (mm.Caption == member || ecaption == member) {
 							type_node = mm;
 							return GetTextFromUrl (mm.URL);
 						}
@@ -714,15 +777,19 @@ public class EcmaHelpSource : HelpSource {
 	{
 		if (!prefix.EndsWith(":"))
 			throw new ArgumentException("prefix");
+
+		if (member != null)
+			member = member.Replace ("#", ".").Replace ("{", "<").Replace ("}", ">");
 			
 		// If a nested type, compare only inner type name to node list.
 		// This should be removed when the node list doesn't lose the containing type name.
 		type = ToEscapedTypeName (type.Replace("+", "."));
+		MatchAttempt[] attempts = GetAttempts (ns, type);
 
 		foreach (Node ns_node in Tree.Nodes){
 			string ns_node_namespace = ns_node.Element.Substring (2);
 
-			if (ns_node_namespace != ns)
+			if (!MatchesNamespace (attempts, ns_node_namespace, out type))
 				continue;
 			
 			foreach (Node type_node in ns_node.Nodes){
@@ -785,25 +852,76 @@ public class EcmaHelpSource : HelpSource {
 		match_node = null;
 		return null;
 	}
+
+	struct MatchAttempt {
+		public string Namespace;
+		public string Type;
+
+		public MatchAttempt (string ns, string t)
+		{
+			Namespace = ns;
+			Type = t;
+		}
+	}
+
+	private static MatchAttempt[] GetAttempts (string ns, string type)
+	{
+		MatchAttempt[] attempts = new MatchAttempt [Count (ns, '.') + 1];
+		int wns = 0;
+		for (int i = 0; i < ns.Length; ++i) {
+			if (ns [i] == '.')
+				attempts [wns++] = new MatchAttempt (ns.Substring (0, i), 
+						ns.Substring (i+1) + "." + type);
+		}
+		attempts [wns++] = new MatchAttempt (ns, type);
+		return attempts;
+	}
+
+	private static int Count (string s, char c)
+	{
+		int n = 0;
+		for (int i = 0; i < s.Length; ++i)
+			if (s [i] == c)
+				++n;
+		return n;
+	}
+
+	private static bool MatchesNamespace (MatchAttempt[] attempts, string ns, out string type)
+	{
+		for (int i = 0; i < attempts.Length; ++i)
+			if (ns == attempts [i].Namespace) {
+				type = attempts [i].Type;
+				return true;
+			}
+		type = null;
+		return false;
+	}
 	
 	public static string ToEscapedTypeName (string typename)
 	{
-		StringBuilder filename = new StringBuilder (typename.Length);
+		return ToEscapedName (typename, "`");
+	}
+
+	static string ToEscapedName (string name, string escape)
+	{
+		StringBuilder filename = new StringBuilder (name.Length);
 		int numArgs = 0;
 		int numLt = 0;
 		bool copy = true;
 
-		for (int i = 0; i < typename.Length; ++i) {
-			char c = typename [i];
+		for (int i = 0; i < name.Length; ++i) {
+			char c = name [i];
 			switch (c) {
+				case '{':
 				case '<':
 					copy = false;
 					++numLt;
 					break;
+				case '}':
 				case '>':
 					--numLt;
 					if (numLt == 0) {
-						filename.Append ('`').Append ((numArgs+1).ToString());
+						filename.Append (escape).Append ((numArgs+1).ToString());
 						numArgs = 0;
 						copy = true;
 					}
@@ -820,6 +938,11 @@ public class EcmaHelpSource : HelpSource {
 		}
 		return filename.ToString ();
 	}
+
+	static string ToEscapedMemberName (string membername)
+	{
+		return ToEscapedName (membername, "``");
+	}
 	
 	public override string GetNodeXPath (XPathNavigator n)
 	{
@@ -870,16 +993,16 @@ public class EcmaHelpSource : HelpSource {
 			match_node = ns_node;
 			string ns_name = nsurl.Substring (2);
 			
-			XmlReader summary = GetNamespaceDocument (ns_name);
-			if (summary == null)
+			XmlDocument doc = GetHelpXmlWithChanges("xml.summary." + ns_name);
+			if (doc == null)
 				return null;
 
 			XsltArgumentList args = new XsltArgumentList();
 			args.AddExtensionObject("monodoc:///extensions", ExtObject);
 			args.AddParam("show", "", "namespace");
 			args.AddParam("namespace", "", ns_name);
-			string s = Htmlize(new XPathDocument (summary), args);
-			return BuildHtml (css_ecma_code, s); 
+			string s = Htmlize(doc, args);
+			return BuildHtml (css_ecma_code, js_code, s); 
 
 		}
 		match_node = null;
@@ -969,6 +1092,23 @@ public class EcmaHelpSource : HelpSource {
 			
 			basetype = basetypedoc.SelectSingleNode("Type/Base/BaseTypeName");
 		}
+		ArrayList extensions = new ArrayList ();
+		foreach (HelpSource hs in RootTree.HelpSources) {
+			EcmaHelpSource es = hs as EcmaHelpSource;
+			if (es == null)
+				continue;
+			Stream s = es.GetHelpStream ("ExtensionMethods.xml");
+			if (s != null) {
+				XmlDocument d = new XmlDocument ();
+				d.Load (s);
+				foreach (XmlNode n in d.SelectNodes ("/ExtensionMethods/*")) {
+					extensions.Add (n);
+				}
+			}
+		}
+		XmlDocUtils.AddExtensionMethods (doc, extensions, delegate (string s) {
+				return RootTree.GetHelpXml ("T:" + s);
+		});
 
 		XsltArgumentList args = new XsltArgumentList();
 
@@ -977,7 +1117,7 @@ public class EcmaHelpSource : HelpSource {
 		if (rest == "") {
 			args.AddParam("show", "", "typeoverview");
 			string s = Htmlize(doc, args);
-			return BuildHtml (css_ecma_code, s); 
+			return BuildHtml (css_ecma_code, js_code, s); 
 		}
 		
 		string [] nodes = rest.Split (new char [] {'/'});
@@ -1026,6 +1166,9 @@ public class EcmaHelpSource : HelpSource {
 		case "O":
 			args.AddParam("membertype", "", "Operator");
 			break;
+		case "X":
+			args.AddParam("membertype", "", "ExtensionMethod");
+			break;
 		case "*":
 			args.AddParam("membertype", "", "All");
 			break;
@@ -1034,7 +1177,7 @@ public class EcmaHelpSource : HelpSource {
 		}
 
 		string html = Htmlize(doc, args);
-		return BuildHtml (css_ecma_code, html); 
+		return BuildHtml (css_ecma_code, js_code, html); 
 	}
 
 	
@@ -1082,11 +1225,12 @@ public class EcmaHelpSource : HelpSource {
 			
 			Stream stream = assembly.GetManifestResourceStream ("mono-ecma-css.xsl");
 			XmlReader xml_reader = new XmlTextReader (stream);
-			ecma_transform_css.Load (xml_reader, null, null);
+			XmlResolver r = new ManifestResourceResolver (".");
+			ecma_transform_css.Load (xml_reader, r, null);
 			
 			stream = assembly.GetManifestResourceStream ("mono-ecma.xsl");
 			xml_reader = new XmlTextReader (stream);
-			ecma_transform_no_css.Load (xml_reader, null, null);
+			ecma_transform_no_css.Load (xml_reader, r, null);
 		}
 		if (use_css)
 			ecma_transform = ecma_transform_css;
@@ -1181,10 +1325,12 @@ public class EcmaHelpSource : HelpSource {
 			return "";
 		}
 
-		public string EditUrlNamespace (string ns, string section)
+		public string EditUrlNamespace (XPathNodeIterator itr, string ns, string section)
 		{
 			if (hs is EcmaUncompiledHelpSource)
 				return "edit:file:" + Path.Combine(((EcmaUncompiledHelpSource)hs).BasePath, ns + ".xml") + "@/Namespace/Docs/" + section; 
+			else if (itr.MoveNext ())
+				return EditingUtils.FormatEditUri(itr.Current.BaseURI, "/elements/" + section);
 			return "";
 		}
 
@@ -1939,7 +2085,7 @@ public class EcmaUncompiledHelpSource : EcmaHelpSource {
 			args.AddExtensionObject("monodoc:///extensions", ExtObject);
 			args.AddParam("show", "", "masteroverview");
 			string s = EcmaHelpSource.Htmlize(new XPathDocument (reader), args);
-			return BuildHtml (css_ecma_code, s); 
+			return BuildHtml (css_ecma_code, js_code, s); 
 		}
 		return base.GetText(url, out match_node);
 	}
